# Copyright 2016 Uri Laserson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from click import group, command, option, argument, File, Choice

from Bio.Alphabet.IUPAC import unambiguous_dna
from Bio.Restriction.Restriction import enzymedict

from pepsyn.operations import reverse_translate
from pepsyn.codons import (
    FreqWeightedCodonSampler, UniformCodonSampler, ecoli_codon_usage)


@group(context_settings={'help_option_names': ['-h', '--help']})
def cli():
    """pepsyn -- peptide synthesis design"""
    pass


# reusable args
argument_input = argument('input', type=File('r'))
argument_output = argument('output', type=File('w'))


@cli.command()
@argument_input
@argument_output
@option('--length', '-l', type=int, help='Length of output oligos')
@option('--overlap', '-p', type=int, help='Overlap of oligos')
def tile(input, output, length, overlap):
    """tile a set of sequences"""
    for seqrecord in SeqIO.parse(input, 'fasta'):
        for (start, end, t) in tile(seqrecord.seq, length, overlap):
                output_title = '{}|{}-{}'.format(seqrecord.id, start, end)
                output_record = SeqRecord(t, output_title)
                SeqIO.write(output_record, output, 'fasta')


@cli.command()
@argument_input
@argument_output
@option('--prefix', '-p', help='DNA sequence to prefix each oligo')
def prefix(input, output, prefix):
    """add a prefix to each sequence"""
    for seqrecord in SeqIO.parse(input, 'fasta'):
        newseq = prefix + seqrecord.seq
        seqrecord.seq = prefix
        SeqIO.write(seqrecord, output, 'fasta')


@cli.command()
@argument_input
@argument_output
@option('--suffix', '-s', help='DNA sequence to suffix each oligo')
def suffix(input, output, suffix):
    """add a suffix to each sequence"""
    for seqrecord in SeqIO.parse(input, 'fasta'):
        newseq = seqrecord.seq + suffix
        seqrecord.seq = newseq
        SeqIO.write(seqrecord, output, 'fasta')


@cli.command()
@argument_input
@argument_output
@option('--codon-table', '-t', default='standard',
        help='ONLY STANDARD TABLE IMPLEMENTED')
@option('--codon-usage', '-u', default='ecoli', help='ONLY ECOLI IMPLEMENTED')
@option('--sampler', default='weighted', show_default=True,
        type=Choice(['weighted', 'uniform']), help='Codon sampling method')
def revtrans(input, codon_table, codon_usage, sampler):
    """reverse translate amino acid sequences into DNA"""
    if sampler == 'weighted':
        codon_sampler = FreqWeightedCodonSampler(usage=ecoli_codon_usage)
    elif sampler == 'uniform':
        codon_sampler = UniformCodonSampler()
    for seqrecord in SeqIO.parse(input, 'fasta'):
        dna_id = seqrecord.id
        dna_seq = reverse_translate(seqrecord.seq, codon_sampler)
        SeqIO.write(SeqRecord(dna_seq, dna_id), output, 'fasta')


@cli.command()
@argument_input
@argument_output
@option('--site', help='Site to remove (e.g., EcoRI, AGCCT); case sensitive')
@option('--start', type=int, help='Start position of CDS (Pythonic coords)')
@option('--end', type=int, help='End position of CDS (Pythonic coords)')
@option('--codon-table', '-t', default='standard',
        help='ONLY STANDARD TABLE IMPLEMENTED')
@option('--codon-usage', '-u', default='ecoli', help='ONLY ECOLI IMPLEMENTED')
@option('--sampler', default='weighted', show_default=True,
        type=Choice(['weighted', 'uniform']), help='Codon sampling method')
def removesite(input, output, site, start, end, codon_table, codon_usage,
               sampler):
    """remove site from each sequence's CDS by recoding"""
    if sampler == 'weighted':
        codon_sampler = FreqWeightedCodonSampler(usage=ecoli_codon_usage)
    elif sampler == 'uniform':
        codon_sampler = UniformCodonSampler()

    if site in enzymedict:
        site = Seq(enzymedict[site]['site'], unambiguous_dna)
    else:
        site = Seq(site, unambiguous_dna)

    for seqrecord in SeqIO.parse(input, 'fasta'):
        id_ = seqrecord.id
        seq = remove_site_from_cds(seqrecord.seq, site, codon_sampler, start,
                                   end)
        SeqIO.write(SeqRecord(seq, id_), output, 'fasta')


@cli.command()
@argument_input
def stats(input):
    """compute some sequence statistics"""
    pass



# Copyright 2016 Uri Laserson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from Bio.Seq import Seq

def tile(seq, length, overlap):
    if length <= 0:
        raise ValueError('length must be a positive integer')
    if overlap < 0:
        raise ValueError('overlap must be a nonnegative integer')
    if overlap >= length:
        raise ValueError('length must be greater than overlap')
    if len(seq) < length:
        raise ValueError('length cannot be larger than the tiled seq')
    end = length
    while end < len(seq):
        start = end - length
        yield (start, end, seq[start:end])
        end += length - overlap
    if end == len(seq):
        start = end - length
        yield (start, end, seq[start:end])
    else:
        start = len(seq) - length
        end = len(seq)
        yield (start, end, seq[start:end])


def reverse_translate(seq, codon_sampler):
    """
    seq is a Bio.Seq.Seq
    codon_sampler is a CodonSampler
    """
    codons = []
    for aa in seq:
        codons.append(codon_sampler.sample_codon(aa))
    return sum(codons, Seq('', codon_sampler.nucleotide_alphabet))


def remove_site_from_cds(seq, site, codon_sampler, cds_start=None,
                         cds_end=None):
    """
    seq is a Bio.Seq.Seq
    site is a Bio.Seq.Seq
    codon_sampler is a CodonSampler
    cds_start, cds_end are int; defaults to full seq
    """
    site_start = seq.find(site)
    if site_start == -1:
        return seq
    site_end = site_start + len(site)
    if cds_start is None:
        cds_start = 0
    if cds_end is None:
        cds_end = len(seq)
    if site_end <= cds_start or site_start >= cds_end:
        # site doesn't overlap the CDS => no-op
        return seq
    if (cds_end - cds_start) % 3 != 0:
        raise PepsynError('CDS length is not multiple of 3')

    # computes offsets for the site boundaries to align them to coding frame
    start_offset = (site_start - cds_start) % 3
    # negative because the end coord must be pushed to the right
    end_offset = -(site_end - cds_start) % 3
    repl_start = max(site_start - start_offset, cds_start)
    repl_end = min(site_end + end_offset, cds_end)

    # compute candidate recoded sequence
    trans = seq[repl_start:repl_end].translate(table=codon_sampler.table)
    candidate = assign_codons(trans, codon_sampler)
    mutseq = seq.tomutable()
    mutseq[repl_start:repl_end] = candidate
    if mutseq.find(site) == -1:
        return mutseq.toseq()
    else:
        # TODO: this can cause a maximum recursion error
        return remove_site_from_cds(
            mutseq.toseq(), site, codon_sampler, cds_start=cds_start,
            cds_end=cds_end)



# Copyright 2016 Uri Laserson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from Bio.Seq import Seq
from Bio.Alphabet.IUPAC import unambiguous_dna
from Bio.Data.CodonTable import standard_dna_table


class CodonUsage(object):

    def __init__(self, freq):
        """freq is a dict of Seq('NNN', alphabet) -> float mappings"""
        # TODO: input verification:
        # all codons
        # sums to 1
        self.freq = freq
        self.nucleotide_alphabet = freq.keys().__iter__().__next__().alphabet


class CodonSampler(object):

    def __init__(self, table=None):
        """
        table is Bio.Data.CodonTable.NCBICodonTableDNA
        """
        self.table = table if table is not None else standard_dna_table
        self.nucleotide_alphabet = self.table.nucleotide_alphabet
        self.protein_alphabet = self.table.protein_alphabet
        # generate aa -> codon mapping
        self.aa2codons = {}
        for (codon, aa) in self.table.forward_table.items():
            c = Seq(codon, self.table.nucleotide_alphabet)
            self.aa2codons.setdefault(aa, []).append(c)

    def sample_codon(self, aa):
        """
        aa is str for single-letter IUPAC AA
        """
        raise NotImplementedError()


class UniformCodonSampler(CodonSampler):

    def __init__(self, table=None):
        """
        table is Bio.Data.CodonTable.NCBICodonTableDNA
        """
        super().__init__(table)

    def sample_codon(self, aa):
        i = np.random.randint(len(self.aa2codons[aa]))
        return self.aa2codons[aa][i]


class FreqWeightedCodonSampler(CodonSampler):

    def __init__(self, table=None, usage=None):
        """
        table is Bio.Data.CodonTable.NCBICodonTableDNA
        usage is CodonUsage; None means uniform
        """
        super().__init__(table)
        if usage is None:
            raise ValueError('Must provide a CodonUsage object')
        if table.nucleotide_alphabet != usage.nucleotide_alphabet:
            raise ValueError('table and usage need to use the same nucleotide'
                             'Alphabet')
        # precalculate distribution (p) for each amino acid
        self.aa2p = {}
        if self.usage is not None:
            for aa in self.table.protein_alphabet.letters:
                unnormed = np.asarray(
                    [self.usage.freq[c] for c in self.aa2codons[aa]])
                self.aa2p[aa] = unnormed / unnormed.sum()

    def sample_codon(self, aa):
        """
        aa is str for single-letter IUPAC AA
        """
        i = np.random.choice(range(len(self.aa2codons[aa])), p=self.aa2p[aa])
        return self.aa2codons[aa][i]


# http://www.kazusa.or.jp/codon/cgi-bin/showcodon.cgi?species=37762
ecoli_codon_usage = CodonUsage({
    Seq('GGG', unambiguous_dna): 12.32,
    Seq('GGA', unambiguous_dna): 13.61,
    Seq('GGT', unambiguous_dna): 23.72,
    Seq('GGC', unambiguous_dna): 20.58,
    Seq('GAG', unambiguous_dna): 19.37,
    Seq('GAA', unambiguous_dna): 35.06,
    Seq('GAT', unambiguous_dna): 33.75,
    Seq('GAC', unambiguous_dna): 17.86,
    Seq('GTG', unambiguous_dna): 19.87,
    Seq('GTA', unambiguous_dna): 13.07,
    Seq('GTT', unambiguous_dna): 21.56,
    Seq('GTC', unambiguous_dna): 13.09,
    Seq('GCG', unambiguous_dna): 21.09,
    Seq('GCA', unambiguous_dna): 23.00,
    Seq('GCT', unambiguous_dna): 18.89,
    Seq('GCC', unambiguous_dna): 21.63,
    Seq('AGG', unambiguous_dna): 3.96,
    Seq('AGA', unambiguous_dna): 7.11,
    Seq('AGT', unambiguous_dna): 13.19,
    Seq('AGC', unambiguous_dna): 14.27,
    Seq('AAG', unambiguous_dna): 15.30,
    Seq('AAA', unambiguous_dna): 37.21,
    Seq('AAT', unambiguous_dna): 29.32,
    Seq('AAC', unambiguous_dna): 20.26,
    Seq('ATG', unambiguous_dna): 23.75,
    Seq('ATA', unambiguous_dna): 13.33,
    Seq('ATT', unambiguous_dna): 29.58,
    Seq('ATC', unambiguous_dna): 19.40,
    Seq('ACG', unambiguous_dna): 13.64,
    Seq('ACA', unambiguous_dna): 15.14,
    Seq('ACT', unambiguous_dna): 13.09,
    Seq('ACC', unambiguous_dna): 18.94,
    Seq('TGG', unambiguous_dna): 13.39,
    Seq('TGA', unambiguous_dna): 1.15,
    Seq('TGT', unambiguous_dna): 5.86,
    Seq('TGC', unambiguous_dna): 5.48,
    Seq('TAG', unambiguous_dna): 0.32,
    Seq('TAA', unambiguous_dna): 2.00,
    Seq('TAT', unambiguous_dna): 21.62,
    Seq('TAC', unambiguous_dna): 11.69,
    Seq('TTG', unambiguous_dna): 12.91,
    Seq('TTA', unambiguous_dna): 17.43,
    Seq('TTT', unambiguous_dna): 24.36,
    Seq('TTC', unambiguous_dna): 13.95,
    Seq('TCG', unambiguous_dna): 8.18,
    Seq('TCA', unambiguous_dna): 13.09,
    Seq('TCT', unambiguous_dna): 13.08,
    Seq('TCC', unambiguous_dna): 9.71,
    Seq('CGG', unambiguous_dna): 7.91,
    Seq('CGA', unambiguous_dna): 4.81,
    Seq('CGT', unambiguous_dna): 15.93,
    Seq('CGC', unambiguous_dna): 14.04,
    Seq('CAG', unambiguous_dna): 26.74,
    Seq('CAA', unambiguous_dna): 14.42,
    Seq('CAT', unambiguous_dna): 12.41,
    Seq('CAC', unambiguous_dna): 7.34,
    Seq('CTG', unambiguous_dna): 37.44,
    Seq('CTA', unambiguous_dna): 5.56,
    Seq('CTT', unambiguous_dna): 14.51,
    Seq('CTC', unambiguous_dna): 9.47,
    Seq('CCG', unambiguous_dna): 14.50,
    Seq('CCA', unambiguous_dna): 9.11,
    Seq('CCT', unambiguous_dna): 9.49,
    Seq('CCC', unambiguous_dna): 6.17})



# Copyright 2016 Uri Laserson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

def readme():
    with open('README.md', 'r') as ip:
        return ip.read()

setup(
    name='pepsyn',
    version='0.0.0',
    description='Peptide library design',
    long_description=readme(),
    url='https://github.com/lasersonlab/pepsyn',
    author='Laserson Lab',
    license='Apache License, Version 2.0',
    classifiers=['Programming Language :: Python :: 3'],
    packages=find_packages(),
    install_requires=['click', 'biopython'],
    entry_points={'console_scripts': ['pepsyn = pepsyn.cli:cli']}
)
